"""
This class generates insert statements based on a mapping.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from typing import List

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference
from stimula.stml.sql.foreign_where_renderer import ForeignWhereClauseRenderer
from stimula.stml.sql.select_renderer import SelectRenderer


class InsertRenderer:
    """
        header: 'c1[unique=true], c2(c1), c3(b1), c4(b2(a1))'

        query:
        insert into c (c1, c2, c3, c4)
        select :1, c.c0, b.b0, b1.b0
        from c, b, b as b0
        join a on b0.b2=a.a0
        where c.c1 = :2 and b.b1 = :3 and a.a1=:4
        returning id;
    """

    def render(self, mapping):
        insert_clause = InsertClauseRenderer().render(mapping)
        select_clause = SelectClauseRenderer().render(mapping)
        from_clause = FromClauseRenderer(True).render(mapping)
        where_clause = ForeignWhereClauseRenderer(True, False).render(mapping)
        returning_clause = ReturningClauseRenderer().render(mapping)

        # this where clause is also used in update query. Here we need 'where'
        if where_clause:
            where_clause = ' where ' + where_clause

        return f'{insert_clause}{select_clause}{from_clause}{where_clause}{returning_clause}'


class InsertClauseRenderer:
    def render(self, mapping: Entity):
        # get attributes. Skip extensions on base table, because they are not columns. We'll insert them in a separate query
        attributes = [a.name for a in mapping.attributes if not (isinstance(a, Reference) and a.extension)]

        # assert there's at least one attribute to insert
        assert attributes, f'No attributes to insert in table {mapping.name}'

        # comma separate
        return f"insert into {mapping.name}({', '.join(attributes)})"


class SelectClauseRenderer:
    def render(self, mapping: Entity):
        # get attributes, skip extensions on base table, because they are not columns. We'll insert them in a separate query
        attributes = [self._attribute(a) for a in mapping.attributes if not (isinstance(a, Reference) and a.extension)]

        # comma separate cells
        return ' select ' + ', '.join(attributes)

    def _attribute(self, attribute: AbstractAttribute):

        if isinstance(attribute, Attribute):
            if attribute.key:
                # build json object
                return f"jsonb_build_object('{attribute.key}', :{attribute.parameter})"

            # get value from parameter
            return f':{attribute.parameter}'

        if isinstance(attribute, Reference):
            # table names may need an alias
            target_alias = attribute.alias or attribute.table
            # no need to recurse
            return f'{target_alias}.{attribute.target_name}'


'''
There's a subtle difference in the from clause between insert and update statements with respect to extensions:
insert query: we must not join with extension table, because we're inserting a new record and we don't have a primary key yet.
update query: we must join with extension table, because we're updating an existing record and we need to filter by the extension table.
'''


class FromClauseRenderer:

    def __init__(self, is_insert_query):
        # if this is an insert query, we must not join with extension tables
        self._is_insert_query = is_insert_query

    def render(self, mapping: Entity):
        # compile and get result as list
        clauses = self.compile_as_list(mapping)

        if not clauses:
            return ''
        return ' from ' + ', '.join(clauses)

    def compile_as_list(self, mapping: Entity)-> List[str]:
        # compile attributes and return as list, so we can also use it to create ORM foreign key queries
        columns = [self._attribute(a, mapping.name) for a in mapping.attributes if isinstance(a, Reference)]

        # filter out empty columns that don't need a from clause
        return [column for column in columns if column]


    def _attribute(self, attribute: Reference, source_alias):

        # if foreign key is an extension on the root table, then we must not join if it's an insert query
        if attribute.extension and self._is_insert_query:
            return None

        # add target table
        from_clause = attribute.table

        # add alias
        target_alias = attribute.alias or attribute.table
        if target_alias != attribute.table:
            from_clause += f' as {target_alias}'

        # recurse
        return from_clause + self._attributes(attribute.attributes, target_alias)

    def _attributes(self, attributes, source_alias):
        # assert len(attributes) == 1, f'Can only insert a single attribute per column, found: {attributes}'
        from_clauses = [self._nested_attribute(a, source_alias) for a in attributes if isinstance(a, Reference)]
        return ''.join([c for c in from_clauses if c])

    def _nested_attribute(self, attribute: Reference, source_alias):

        # assume we need a left join. This may require a modifier at some point.
        from_clause = f' left join {attribute.table}'

        # add alias
        target_alias = attribute.alias or attribute.table
        if target_alias != attribute.table:
            from_clause += f' as {target_alias}'

        # add on clause
        from_clause += f' on {source_alias}.{attribute.name} = {target_alias}.{attribute.target_name}'

        # if this is an Odoo style extension relation and it's not on the root table, then we need to filter by qualifier (module) and table (model)
        if attribute.extension:
            # assume for now that source alias is the table name. This is fine as long as we're not joining the same table multiple times
            table_name = SelectRenderer().get_model_name(source_alias)
            from_clause += f' and {target_alias}.model = \'{table_name}\' and {target_alias}.module = \'{attribute.qualifier}\''

        # recurse
        return from_clause + self._attributes(attribute.attributes, target_alias)


class ReturningClauseRenderer:
    # returns 'returning id' if this mapping contains an extension relation on the root table

    def render(self, mapping: Entity):
        # get list of returned columns
        clauses = [self._attribute(a) for a in mapping.attributes]

        # filter out all empty clauses
        non_empty_clauses = [c for c in clauses if c]

        if not any(non_empty_clauses):
            return ''

        # assert there's no more than one
        assert len(non_empty_clauses) == 1, f'Can only return a single extension id, found: {non_empty_clauses}'

        return ' returning ' + mapping.name + '.' + non_empty_clauses[0]

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            # not a foreign key relationship, so not an extension
            return None

        if isinstance(attribute, Reference):
            if not attribute.extension:
                # not an extension foreign key
                return None
            # extension, so return the id to return from the query. Default to 'id' bec/ that's what Odoo uses
            return attribute.id or 'id'



"""
This class is used to execute queries in a transactional way. It will execute the queries in rounds until no new successful queries are found.

Author: Romke Jonker
Email: romke@stml.io
"""

import logging

from stimula.service.context import cnx_context
from stimula.service.query_executor import OperationType

_logger = logging.getLogger(__name__)


class ExecutorService:
    def execute_sql(self, query_executors, execute, commit, tx_size=1000):
        if not execute:
            # fake execution, return result
            return [qe.fake_execute() for qe in query_executors]

        # get cursor from context
        cr = cnx_context.cr

        # execute queries, rerun until exhausted
        result = self._eat_sleep_repeat(query_executors, cr, commit, tx_size)

        # commit if requested
        if commit:

            # registry may not be available during unit tests
            if hasattr(cnx_context, 'registry'):
                # registry may not have clear_cache() method
                if hasattr(cnx_context.registry, 'clear_cache'):
                    # invalidate caches to avoid stale values coming from cache
                    cnx_context.registry.clear_cache()
                else:
                    _logger.warning("Registry has no clear_cache() method")

        return result

    def _eat_sleep_repeat(self, query_executors, cr, commit, tx_size):
        # execute in rounds until no new successful queries are found

        # create result lists
        completed = []
        failed = []

        # copy query executors list
        remaining = query_executors.copy()

        done = False
        tx_count = 0

        while not done:
            new_completed_executors = []
            new_completed_results = []
            # iterate query executors
            for query_executor in remaining:
                # create or replace savepoint
                self.create_savepoint()
                # delegate execution to query executor
                execution_result = query_executor.execute(cr)
                # if successful
                if execution_result.success:
                    # append result to list
                    new_completed_results.append(execution_result)
                    # remove from remaining
                    new_completed_executors.append(query_executor)
                    # increment tx count
                    tx_count += 1
                    if tx_count >= tx_size:
                        # commit transaction
                        if commit:
                            cnx_context.cnx.commit()
                        else:
                            # rollback all queries
                            cnx_context.cnx.rollback()
                        # reset tx count
                        tx_count = 0
                else:
                    # rollback to savepoint
                    self.rollback_to_savepoint()
                    # append to failed list
                    failed.append(execution_result)
            if new_completed_results:
                # append new completed to completed list
                completed.extend(new_completed_results)
                # remove completed executors from remaining
                remaining = [qe for qe in remaining if qe not in new_completed_executors]
                # reset failed list and start again
                failed = []
            else:
                # nothing new completed, we're done
                done = True
                # commit if transactions remain
                if tx_count > 0 and commit:
                    cnx_context.cnx.commit()

        # combine completed and failed lists
        all_results = completed + failed

        # set delete queries apart, because they don't have line numbers
        deleted = [result for result in all_results if result.operation_type == OperationType.DELETE]
        insert_and_updates = [result for result in all_results if result.operation_type != OperationType.DELETE]

        # sort by line_number
        insert_and_updates.sort(key=lambda x: x.line_number)

        # append deleted to the end
        return insert_and_updates + deleted

    def create_savepoint(self):
        # create savepoint
        cnx_context.cr.execute("SAVEPOINT stimula_savepoint")

    def rollback_to_savepoint(self):
        # rollback to savepoint
        cnx_context.cr.execute("ROLLBACK TO SAVEPOINT stimula_savepoint")



"""
This class compiles a mapping into a dictionary of parameter types for each parameter.

Author: Romke Jonker
Email: romke@stml.io
"""
from itertools import chain
from typing import Tuple, List

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class ParameterTypesRenderer:

    def render(self, mapping: Entity):
        # compile columns into a list of dictionaries
        attributes = self._attributes(mapping.attributes)

        # combine the dictionaries into one
        return {a[0]: a[1] for a in attributes}

    def _attributes(self, attributes: List[AbstractAttribute]) -> List[Tuple[str, str]]:
        return list(chain(*[self._attribute(a) for a in attributes]))

    def _attribute(self, attribute: AbstractAttribute) -> List[Tuple[str, str]]:
        if isinstance(attribute, Attribute):
            return [(attribute.parameter, attribute.type)]

        if isinstance(attribute, Reference):
            return self._attributes(attribute.attributes)



"""
This class compiles a mapping into a select query.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import re
from itertools import chain
from typing import List, Tuple

from stimula.stml.alias_enricher import AliasEnricher
from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class SelectRenderer:
    """
        select c.c1, c0.c1, b.b1, a.a1
        from c
        left join c as c0 on c.c2 = c0.c0
        left join b on c.c3 = b.b0
        left join b as b0 on c.c4 = b0.b0
        left join a on b0.b2=a.a0
        order by c.c1
    """

    def render(self, mapping: Entity, where_clause=None):
        """
        Compiles a mapping into a select query
        :param mapping: the mapping
        :param where_clause: a free where clause for the caller to specify
        :return: the select query
        """
        select_clause = SelectClauseRenderer().render(mapping)
        join_clause = JoinClauseRenderer().render(mapping)
        order_by_clause = OrderByClauseRenderer().render(mapping)

        # render [filter="...$..."] headers into a where clause, replacing '$' with the column name
        where = WhereClauseRenderer().render(mapping, where_clause)
        return f'{select_clause} from {mapping.name}{join_clause}{where}{order_by_clause}'

    def compile_count_query(self, mapping: Entity, where_clause=None):
        table = mapping.name
        aliased_mapping = AliasEnricher().enrich(mapping)
        select_clause = 'select count(*)'
        join_clause = JoinClauseRenderer().render(aliased_mapping)
        where = f' where {where_clause}' if where_clause else ''
        return f'{select_clause} from {table}{join_clause}{where}'

    def get_model_name(self, alias):
        # this method gets the model name to use in the extension clause
        # current solution is based on the alias, which is not very robust
        # alias is of the form 'xxx_xxx_xxx_d', we need to convert that into 'xxx.xxx.xxx'

        # remove '_0' postfix from alias using regex if it exists
        alias = re.sub(r'_\d+$', '', alias)
        return alias.replace('_', '.')


class SelectClauseRenderer:
    def render(self, mapping: Entity):

        # Include empty columns. Skip cells with skip=true or orm-only modifier. We need those when reading CSV, but not when reading from DB
        attributes = [self._attribute(a, mapping.name) for a in mapping.attributes if (not a) or (not a.skip and not a.orm_only)]

        # join attributes per column
        joined_columns = [self._join_attributes(a) for a in attributes]

        # comma separate columns
        return 'select ' + ', '.join(joined_columns)

    def _attribute(self, attribute: AbstractAttribute, alias) -> List[Tuple]:
        # column may be empty
        if not attribute:
            return [('\'\'',)]

        if isinstance(attribute, Attribute):
            # if there's a key, then return the json field
            if attribute.key:
                return [(f"{alias}.{attribute.name}->>'{attribute.key}'", attribute)]

            # return alias and column name. Also return the source attribute, so we can check the type later
            return [(f'{alias}.{attribute.name}', attribute)]

        if isinstance(attribute, Reference):
            # table names may need an alias
            target_alias = attribute.alias or attribute.table

            # recurse into foreign key table to get the attributes
            return self._nested_attributes(attribute.attributes, target_alias)

    def _nested_attributes(self, attributes: List[AbstractAttribute], alias: str) -> List[Tuple]:

        # flatten the attributes for this column
        return list(chain(*[self._attribute(p, alias) for p in attributes]))

    def _join_attributes(self, selects: List[Tuple]) -> 'str':
        # if there's more than one attribute, we must cast jsonb columns to string
        if len(selects) > 1:
            selects = self._cast_selects(selects)

        # get names
        names = [p[0] for p in selects]

        # colon separate attributes
        return ' || \':\' || '.join(names)

    def _cast_selects(self, selects):
        # postgres can't concatenate jsonb columns to anything, so cast to string
        cast_selects = []
        for select in selects:
            # select is a tuple of (column, attribute)
            if select[1].type == 'jsonb':
                cast_selects.append((f'{select[0]}::text', select[1]))
            else:
                cast_selects.append(select)
        return cast_selects


class JoinClauseRenderer:
    def render(self, mapping: Entity):
        table_name = mapping.name

        # get join clauses
        clauses = list(chain(*[self._attribute(a, table_name, a.unique) for a in mapping.attributes if a]))

        # glue cells together, skip empty columns
        return ''.join(clauses)

    def _attribute(self, attribute: AbstractAttribute, alias, unique: bool) -> List[str]:
        # if no foreign key, return empty string
        if isinstance(attribute, Attribute):
            return []

        if isinstance(attribute, Reference):
            # but table names may need an alias
            source_name = attribute.name
            target_name = attribute.target_name
            target_table = attribute.table
            target_alias = attribute.alias or target_table

            # Assume we need a left join. This may require a modifier at some point.
            join_clause = f' left join {target_table}'

            # only add 'as alias' if it's different from table name
            if target_alias != target_table:
                join_clause += f' as {target_alias}'
            join_clause += f' on {alias}.{source_name} = {target_alias}.{target_name}'

            # if this is an Odoo style extension relation, then we need to filter by qualifier (module) and table (model)
            if attribute.qualifier:
                if unique:
                    # for an extension table in a unique attribute, we should do an inner join instead of a left join, because we want to skip records that don't have an extension record
                    join_clause = join_clause.replace(' left join ', ' join ')

                # assume for now that alias is the table name. This is fine as long as we're not joining the same table multiple times
                table_name = SelectRenderer().get_model_name(alias)
                join_clause += f' and {target_alias}.model = \'{table_name}\' and {target_alias}.module = \'{attribute.qualifier}\''

            # recurse
            return [join_clause] + self._attributes(attribute.attributes, target_alias, unique)

    def _attributes(self, attributes: List[AbstractAttribute], alias, unique: bool) -> List[str]:
        return list(chain(*[self._attribute(attribute, alias, unique) for attribute in attributes]))


class OrderByClauseRenderer:
    def render(self, mapping: Entity):
        table_name = mapping.name
        # clauses to order by, skip empty columns
        clauses = [self._attribute(a, table_name) for a in mapping.attributes if a and a.unique]

        # nothing to order by
        if not clauses:
            return ''

        return ' order by ' + ', '.join(clauses)

    def _attribute(self, attribute: AbstractAttribute, alias: str):
        if isinstance(attribute, Attribute):
            # if no foreign key, get value from parameter
            return f'{alias}.{attribute.name}'

        if isinstance(attribute, Reference):
            # but table names may need an alias
            attributes = attribute.attributes
            target_name = attribute.table
            target_alias = attribute.alias or target_name
            # recurse
            return ', '.join([self._attribute(p, target_alias) for p in attribute.attributes])


class WhereClauseRenderer:
    def render(self, mapping: Entity, where_clause):
        """
        Compiles [select="...$..."] modifiers into a where clause
        :param mapping:
        :param where_clause: additional where clause to append
        :return: the where clause
        """
        table_name = mapping.name

        # get conditions from columns with filter modifiers, skip empty columns
        conditions = list(chain(*[self._attribute(a, table_name) for a in mapping.attributes if a]))

        # append user-defined where clause
        if where_clause:
            conditions.append(where_clause)

        if not conditions:
            return ''

        # return where clause with conditions joined by ' and '
        return ' where ' + ' and '.join(conditions)

    def _attribute(self, attribute, alias) -> List[str]:
        # if this is not a foreign key, return alias and column name
        if isinstance(attribute, Attribute):
            if not attribute.filter:
                return []

            # also return the source attribute, so we can check the type later
            name = f'{alias}.{attribute.name}'
            return [attribute.filter.replace('$', name)]

        if isinstance(attribute, Reference):
            # table names may need an alias
            target_alias = attribute.alias or attribute.target_name
            # recurse into foreign key table to get the attributes
            return self._attributes(attribute.attributes, target_alias)

    def _attributes(self, attributes: List[AbstractAttribute], alias):
        # flatten list
        return list(chain(*[self._attribute(p, alias) for p in attributes]))



"""
This class generates the query needed to resolved foreign keys for the ORM insert operation.

Author: Romke Jonker
Email: romke@stml.io
"""
from itertools import chain
from typing import List

from stimula.stml.sql.foreign_where_renderer import ForeignWhereClauseRenderer
from stimula.stml.sql.insert_renderer import FromClauseRenderer
from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class OrmInsertRenderer:
    """
        header: 'c1[unique=true], c2(c1), c3(b1), c4(b2(a1))'

        query:
        select c.c0, b.b0, b1.b0
        from c, b, b as b0
        join a on b0.b2=a.a0
        where c.c1 = :2 and b.b1 = :3 and a.a1=:4
    """

    def render(self, mapping: Entity):
        select_clause = SelectClauseRenderer().render(mapping)
        from_clause = FromClauseRenderer(False).render(mapping)
        where_clause = ForeignWhereClauseRenderer(False, False).render(mapping)

        # this where clause is also used in update query. Here we need 'where'
        if where_clause:
            where_clause = ' where ' + where_clause

        return f'{select_clause}{from_clause}{where_clause}'


class SelectClauseRenderer:
    def render(self, mapping: Entity):
        # get list of attributes
        attributes = self._attributes(mapping.attributes)

        # comma separate cells. For an insert, there may be nothing to select
        return 'select ' + ', '.join(attributes) if attributes else ''


    def _attributes(self, attributes: List[AbstractAttribute]):
        # iterate attributes to get list of lists.
        return list(chain(*[self._attribute(a) for a in attributes]))

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            # nothing to select for simple parameters
            return []

        if isinstance(attribute, Reference):
            # table names may need an alias
            # TODO: check if we need to alias the target column
            target_alias = attribute.alias or attribute.table
            # no need to recurse
            return [f'{target_alias}.{attribute.target_name} as {attribute.name}']



class OrmParameterNamesRenderer:
    # returns a list of parameter names that we need to pass to the orm create function, excluding
    # those that come from the query
    def render(self, mapping: Entity):
        # get list of attributes
        return self._attributes(mapping.attributes)

    def _attributes(self, attributes: List[AbstractAttribute]):
        # iterate attributes and flatten
        return list(chain(*[self._attribute(a) for a in attributes]))

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            # return the parameter name, but without the colon
            return [f'{attribute.parameter}']

        if isinstance(attribute, Reference):
            # we'll resolve foreign keys in they query
            return []




"""
This class returns the dtypes and converters for columns that require special handling.

For a column, it may add a read_csv_converter that pandas can use to read the column from a CSV file into a data frame.
This typically happens when the user posts table contents.

For a column, it may add a write_csv_converter that pandas can use to write the column from a data frame to a CSV file.
This typically happens when the user requests table contents.


Author: Romke Jonker
Email: romke@rnadesign.net
"""
import json
import re
from itertools import chain

import numpy as np
import pandas as pd

from stimula.stml.model import AbstractAttribute, Attribute, Reference


class TypesRenderer:

    def render(self, mapping, column_names, include_skip=False, include_orm_only=False, substitutions=None):
        # only return enabled columns
        if not mapping.attributes:
            return {}

        # process all columns to obtain the converters, include empty columns, but skip [skip=true] columns unless otherwise requested
        attributes = [self._column(a, substitutions) for a in mapping.attributes if (not a) or ((include_skip or not a.skip) and (include_orm_only or not a.orm_only))]

        # create a dictionary of converters to read from csv
        read_csv_converters = {column_names[i]: column['read_csv_converter'] for i, column in enumerate(attributes) if 'read_csv_converter' in column}

        # create a dictionary of converters to write to csv
        write_csv_converters = {column_names[i]: column['write_csv_converter'] for i, column in enumerate(attributes) if 'write_csv_converter' in column}

        # create a dictionary of converters to read from db
        read_db_converters = {column_names[i]: column['read_db_converter'] for i, column in enumerate(attributes) if 'read_db_converter' in column}

        # create a dictionary of dtypes for read_csv
        read_csv_dtypes = {column_names[i]: column['read_csv_dtype'] for i, column in enumerate(attributes) if 'read_csv_dtype' in column}

        # get the columns that must be parsed as dates
        read_csv_parse_dates = [column_names[i] for i, column in enumerate(attributes) if column.get('read_csv_parse_dates', False)]

        # return the dictionary of converters and dtypes
        result = {
            'read_csv_converters': read_csv_converters,
            'write_csv_converters': write_csv_converters,
            'read_db_converters': read_db_converters,
            'read_csv_dtypes': read_csv_dtypes,
            'read_csv_parse_dates': read_csv_parse_dates}

        return result

    def _column(self, attribute: AbstractAttribute, substitutions: str):
        # empty column does not have attributes
        if not attribute:
            return {}

        # convert attribute or reference to list of attributes
        attributes = self._attribute(attribute)

        result = {}

        # to read json from csv, we need to convert the string to json
        if len(attributes) == 1 and attributes[0].type == 'jsonb':
            result['read_csv_converter'] = json_to_dict
            result['write_csv_converter'] = dict_to_json

        # to read binary string from csv, we need to convert the string to binary
        if len(attributes) == 1 and attributes[0].type == 'bytea':
            result['read_csv_converter'] = binary_string_converter

        # to read binary string from db, we need to convert the binary to string
        if len(attributes) == 1 and attributes[0].type == 'bytea':
            result['read_db_converter'] = memoryview_to_string_converter

        # set whether to parse the column as a date
        if self._date_type(attributes):
            result['read_csv_parse_dates'] = True

        # convert date to pandas datetime when reading from db
        if len(attributes) == 1 and attributes[0].type == 'date':
            result['read_db_converter'] = date_to_datetime_converter

        # get the dtype for this column
        dtype = self._dtype(attributes)

        # create list of read_csv_converters to handle multiple converters
        read_csv_converters = []

        # set converter if column has a default value, so that pandas can set it right when reading the posted CSV
        if attribute.default_value:
            read_csv_converters.append(default_value_converter(dtype, attribute.default_value))

        # set substitution converter if column has a substitution. Substitutions are not used when reading from DB
        if attribute.substitute and substitutions:
            read_csv_converters.append(_substitute_converter(substitutions, attribute.substitute))

        # if type is text and there's no read_csv_converter yet, the default is to strip trailing spaces
        if dtype == 'string' and 'read_csv_converter' not in result:
            read_csv_converters.append(strip_trailing_spaces)

        # if there is at least one converter, set the pipeline converter
        if read_csv_converters:
            result['read_csv_converter'] = _pipeline_converter(read_csv_converters)

        # set the dtype
        result['read_csv_dtype'] = dtype

        return result

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            return [attribute]

        if isinstance(attribute, Reference):
            # recurse
            return list(chain(*[self._attribute(a) for a in attribute.attributes]))

    def _dtype(self, attributes):
        # if there are multiple types, return 'string'
        if len(attributes) > 1:
            return 'string'

        # get attribute type
        type = attributes[0].type

        # read boolean as boolean
        if type == 'boolean':
            return 'boolean'

        # read integer as Int64
        if type == 'integer':
            return 'Int64'

        # read numeric as Float64
        if type == 'numeric':
            return 'float'

        # replace timestamp and date columns with 'object', because pandas doesn't like datetime64 columns when reading from csv
        if type == 'timestamp' or type == 'date':
            return 'object'

        # read jsonb as object
        if type == 'jsonb':
            return 'object'

        # double precision as float
        if type == 'double precision':
            return 'float'

        # binary string as object
        if type == 'bytea':
            return 'object'

        # default to string
        return 'string'

    def _date_type(self, attributes):
        # return True if there's a single attribute with type 'date' or 'timestamp'
        if len(attributes) == 0:
            return False
        attribute_type = attributes[0].type
        return attribute_type == 'date' or attribute_type == 'timestamp'


def strip_trailing_spaces(value):
    # strip spaces from string values
    return value.strip() if isinstance(value, str) else value


def json_to_dict(json_str):
    # accept json strings that come from CSV using single quotes.

    # accept empty string, return None
    if not json_str:
        return None

    # get index of first double quote
    first_double_quote = json_str.find('"')

    # get index of first single quote
    first_single_quote = json_str.find("'")

    # if there are no double quotes, or if there are double quotes but there's a single quote that comes before the first double quote
    if first_double_quote == -1 or (0 <= first_single_quote < first_double_quote):
        # escape double quotes
        json_str = json_str.replace('"', '\\"')
        json_str = json_str.replace('\n', '\\n')
        # and replace all single quotes with double quotes
        json_str = json_str.replace("'", '"')

    # parse json string
    try:
        # convert json string into dictionary
        return json.loads(json_str)
    except json.JSONDecodeError as e:
        raise json.JSONDecodeError(f"Error parsing JSON string: {json_str}", e.doc, e.pos) from None


def dict_to_json(dict):
    return json.dumps(dict, ensure_ascii=False)


# Custom converter function to handle binary string data
def binary_string_converter(mixed_string):
    # handle empty string as None, because that's how we probably mean to store it in the DB
    if not mixed_string:
        return None

    # Define a regular expression pattern to match the escaped sequences
    escaped_pattern = re.compile(r'(\\x[0-9a-fA-F]{2})+')

    # Define a function to decode the escaped sequences
    def decode_escaped(match):
        # Extract the hexadecimal characters and decode them using UTF-8
        hex = match.group(0).replace('\\x', '')
        try:
            # try to decode as utf-8
            return bytes.fromhex(hex).decode('utf-8')
        except:
            try:
                # if it fails, try to decode as latin-1
                return bytes.fromhex(hex).decode('latin-1')
            except:
                # if it fails, return the hex string
                return match.group(0)

    # Replace the escaped sequences with their decoded counterparts
    decoded_string = escaped_pattern.sub(decode_escaped, mixed_string)

    return decoded_string


def memoryview_to_string_converter(mv):
    # return None for empty values
    if mv is None:
        return None

    # Function to convert memoryview objects to strings
    try:
        # try to decode as utf-8
        return mv.tobytes().decode('utf-8')
    except UnicodeDecodeError:
        try:
            # try to decode as latin-1
            return mv.tobytes().decode('latin-1')
        except UnicodeDecodeError:
            # if all fails, return as binary
            return mv.tobytes()


def date_to_datetime_converter(date):
    # return None for empty values
    if date is None:
        return None

    # convert date from database to pandas datetime, so that we can accurately compare it to a date read from CSV
    return pd.to_datetime(date)


# Custom converter function to set default values when reading CSV
def default_value_converter(dtype, default):
    # pandas ignores dtype if a converter is set, so we need to convert the value to the correct type
    return lambda value: _convert_to_dtype(dtype, _set_default_value(value, default))


def _set_default_value(value, default):
    return value if value is not None and value != '' and not (isinstance(value, (float, np.float64)) and np.isnan(value)) else default


def _substitute_converter(substitutions, domain):
    return lambda value: _substitute(substitutions, domain, value)


def _substitute(substitutions, domain, value):
    # return None for empty values
    if value is None:
        return None

    # if domain doesn't exist, then throw an error
    assert domain in substitutions, f"Domain '{domain}' not found in substitutions"

    # find a full match
    if value in substitutions[domain]:
        return substitutions[domain][value]

    # then try regular expressions
    for key, substitution in substitutions[domain].items():
        if re.match(key, value):
            return substitution

    # else return original value
    return value


def _key_to_frozenset_converter(dtype, key):
    # a dict is not hashable and can't be used as index field, so return frozen set instead of dict when reading from CSV
    return lambda value: frozenset({(key, value)})


def _dict_to_frozenset_converter(dtype):
    # a dict is not hashable and can't be used as index field, so return frozen set instead of dict when reading from DB
    return lambda value: _convert_dict_to_frozenset(dtype, value)


def _convert_dict_to_frozenset(dtype, value):
    if value is None:
        return None
    return frozenset(value.items())


def _convert_to_dtype(dtype, value):
    if dtype == 'boolean':
        return isinstance(value, str) and value.lower() == 'true'

    if dtype == 'Int64':
        return int(value)

    if dtype == 'float':
        return float(value)

    return value


def _pipeline_converter(converters: list):
    return lambda value: _pipeline(value, converters)


def _pipeline(value, converters: list):
    for converter in converters:
        value = converter(value)
    return value



'''
This class expands placeholders in a mapping. A placeholder may expand into multiple values.
The result of expansion is therefore a list of mappings, one for each combination of values.

author: Romke Jonker
email: romke@stml.io
'''
import copy
import itertools
import re
from itertools import chain

from stimula.stml.model import Entity, Attribute, Reference


class ParameterExpander:

    def expand(self, mapping: Entity, substitutions: dict) -> list[Entity]:

        # get placeholders
        placeholders = self._get_placeholders(mapping)

        # if there are no placeholders, return the mapping as is
        if not placeholders:
            return [mapping]

        # get values for all placeholders
        values = {p: list(substitutions.get(p, {}).keys()) for p in placeholders}

        # verify that each placeholder has at least one value
        empty_placeholders = [p for p, v in values.items() if not v]
        if empty_placeholders:
            raise ValueError(f'No value found for placeholders: {empty_placeholders}')

        # create cartesian product of values
        values_product = list(itertools.product(*values.values()))

        # create a list of dicts mapping placeholders to values
        value_maps = [{p: v for p, v in zip(values.keys(), value)} for value in values_product]

        # expand the mapping for each value map
        mappings = [self._expand_values(mapping, value_map) for value_map in value_maps]

        return mappings

    def _get_placeholders(self, mapping):
        # regex to map ${...} placeholders
        regex = r'\$\{([^\}]+)\}'

        #         get list of attributes that can contain a placeholder
        values = list(chain(*[[a.default_value, a.exp, a.key] for a in mapping.attributes if isinstance(a, Attribute)]))
        values.extend(chain(*[[a.default_value, a.exp] for a in mapping.attributes if isinstance(a, Reference)]))

        # find all placeholders in the values
        placeholders = []
        for value in values:
            if value is not None:
                placeholders.extend(re.findall(regex, value))

        return placeholders

    def _expand_values(self, mapping, value_map):
        # create a deep copy of the mapping
        mapping_copy = copy.deepcopy(mapping)

        for a in mapping_copy.attributes:
            if isinstance(a, Attribute):
                a.default_value = self._replace_placeholders(a.default_value, value_map)
                a.exp = self._replace_placeholders(a.exp, value_map)
                a.key = self._replace_placeholders(a.key, value_map)
            if isinstance(a, Reference):
                a.default_value = self._replace_placeholders(a.default_value, value_map)
                a.exp = self._replace_placeholders(a.exp, value_map)

        return mapping_copy

    def _replace_placeholders(self, value, value_map):
        for k, v in value_map.items():
            if value is not None:
                value = value.replace(f'${{{k}}}', v)
        return value

#



from abc import ABC
from typing import List, Dict


class Entity:
    def __init__(self, name: str, attributes: List['AbstractAttribute'] = [], primary_key: str = None) -> None:
        self.name: str = name
        self.primary_key: str = primary_key
        self.attributes: List[AbstractAttribute] = attributes

    def to_dict(self) -> Dict[str, any]:
        data = {
            "name": self.name,
            "primary_key": self.primary_key,
            "attributes": [prop.to_dict() if prop else None for prop in self.attributes]
        }
        return {key: value for key, value in data.items() if value}

    def __repr__(self) -> str:
        return str(self.to_dict())  # Use to_dict for the repr method

    def __eq__(self, other: object) -> bool:
        # Check if other is an instance of Entity and compare relevant attributes
        if not isinstance(other, Entity):
            return False
        return self.name == other.name \
            and _none_safe_compare(self.primary_key, other.primary_key) \
            and self.attributes == other.attributes


class AbstractAttribute(ABC):
    def __init__(self, name: str, unique: bool, skip: bool, exp: str, default_value: str, orm_only: bool, enabled: bool, primary_key: bool, in_use: bool, default: bool, deduplicate: bool,
                 substitute: str, key: str) -> None:
        self.name: str = name
        self.unique: bool = unique
        self.skip: bool = skip
        self.exp: str = exp
        self.default_value: str = default_value
        self.orm_only: bool = orm_only
        self.enabled: bool = enabled
        self.primary_key: bool = primary_key
        self.in_use: bool = in_use
        self.default: bool = default
        self.deduplicate: bool = deduplicate
        self.substitute: str = substitute
        self.key: str = key

    def to_dict(self) -> Dict[str, any]:
        data = {"name": self.name, 'unique': self.unique, 'skip': self.skip, 'exp': self.exp, 'default_value': self.default_value, 'orm_only': self.orm_only, 'enabled': self.enabled,
                'primary_key': self.primary_key, 'in_use': self.in_use, 'default': self.default, 'deduplicate': self.deduplicate, 'substitute': self.substitute, 'key': self.key}
        return {key: value for key, value in data.items() if value}

    def __repr__(self) -> str:
        return str(self.to_dict())  # Use to_dict for the repr method

    def __eq__(self, other: object) -> bool:
        # Check if other is an instance of Attribute and compare the name
        if not isinstance(other, AbstractAttribute):
            return False
        return self.name == other.name \
            and self.unique == other.unique \
            and self.skip == other.skip \
            and self.exp == other.exp \
            and self.default_value == other.default_value \
            and self.orm_only == other.orm_only \
            and self.enabled == other.enabled \
            and self.primary_key == other.primary_key \
            and self.in_use == other.in_use \
            and self.default == other.default \
            and self.deduplicate == other.deduplicate \
            and self.substitute == other.substitute \
            and self.key == other.key


class Attribute(AbstractAttribute):
    def __init__(self, name: str, unique=False, skip=False, exp='', default_value=None, orm_only=False, enabled: bool = False, primary_key: bool = False, in_use: bool = False, default: bool = False,
                 deduplicate: bool = False, substitute: str = None,
                 key: str = None, type: str = None, parameter: str = None, filter: str = None, api: str = None, url: str = None) -> None:
        super().__init__(name, unique, skip, exp, default_value, orm_only, enabled, primary_key, in_use, default, deduplicate, substitute, key)
        self.type: str = type
        self.parameter: str = parameter
        self.filter: str = filter
        self.api: str = api
        self.url: str = url

    def to_dict(self) -> Dict[str, any]:
        data = super().to_dict() | {"type": self.type, "parameter": self.parameter, "filter": self.filter, "api": self.api, "url": self.url}
        return {key: value for key, value in data.items() if value}

    def __eq__(self, other: object) -> bool:
        # Check if other is an instance of Attribute and compare the name
        if not isinstance(other, Attribute):
            return False
        return super().__eq__(other) \
            and _none_safe_compare(self.type, other.type) \
            and _none_safe_compare(self.parameter, other.parameter) \
            and _none_safe_compare(self.filter, other.filter) \
            and _none_safe_compare(self.api, other.api) \
            and _none_safe_compare(self.url, other.url)


class Reference(AbstractAttribute):
    def __init__(self, name: str, attributes: List['AbstractAttribute'] = [], unique: bool = False, skip: bool = False, exp: str = '', default_value: str = None, orm_only: bool = False,
                 enabled: bool = False, primary_key: bool = False, in_use: bool = False, default: bool = False, deduplicate: bool = False, substitute: str = None, key: str = None,
                 table: str = None, target_name: str = None, qualifier: str = None, extension: bool = False, alias: str = None, id: str = None) -> None:
        super().__init__(name, unique, skip, exp, default_value, orm_only, enabled, primary_key, in_use, default, deduplicate, substitute, key)
        self.attributes: List[AbstractAttribute] = attributes
        self.table = table
        self.target_name = target_name
        self.qualifier = qualifier
        self.extension = extension
        self.alias = alias
        self.id = id

    def to_dict(self) -> Dict[str, any]:
        data = super().to_dict() | {"table": self.table, "target_name": self.target_name, "qualifier": self.qualifier, "extension": self.extension, "alias": self.alias, "id": self.id}
        data['attributes'] = [a.to_dict() for a in self.attributes]
        return {key: value for key, value in data.items() if value}

    def __eq__(self, other: object) -> bool:
        # Check if other is an instance of Attribute and compare the name
        if not isinstance(other, Reference):
            return False
        return super().__eq__(other) \
            and self.attributes == other.attributes \
            and _none_safe_compare(self.table, other.table) \
            and _none_safe_compare(self.target_name, other.target_name) \
            and _none_safe_compare(self.qualifier, other.qualifier) \
            and self.extension == other.extension \
            and _none_safe_compare(self.alias, other.alias) \
            and _none_safe_compare(self.id, other.id)


def _none_safe_compare(a, b) -> bool:
    if a is None:
        return b is None
    return a == b



'''
this class is responsible for executing ORM queries

Author: Romke Jonker
Email: romke@stml.io
'''
from abc import abstractmethod

import pandas as pd

from stimula.service.abstract_orm import AbstractORM
from stimula.service.query_executor import Executor, ExecutionResult


class OrmExecutor(Executor):
    def __init__(self, line_number, operation_type, table, query, query_values, orm_values, context, orm):
        super().__init__(line_number, operation_type, table, context)
        self._query: str = query
        self._query_values: dict = query_values
        self._orm_values: dict = orm_values
        self._orm: AbstractORM = orm

    def execute(self, cursor):

        key_values = {}

        if self._query:
            try:
                # execute query
                key_values = self._execute_select_query(cursor, self._query, self._query_values)
            except Exception as e:
                error = str(e)
                return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self._query, self._query_values, self.context, error=error)


        # combine orm values and key values into a new dictionary
        params = {**self._orm_values, **key_values}

        # execute ORM method
        return self._execute_orm_method(self.table_name, params)

    def _execute_select_query(self, cursor, query: str, params: dict):
        # replace ':' style place holders with '%' style
        psycopg_query = self._replace_placeholders(self._query)
        # replace NA values with None in params dictionary
        params_with_none = {k: None if pd.isna(v) else v for k, v in self._query_values.items()}
        cursor.execute(psycopg_query, params_with_none)
        row = cursor.fetchone()
        rowcount = cursor.rowcount

        # verify row was affected
        if rowcount == 0:
            raise ValueError("No row was affected. Query: %s, Params: %s" % (query, params))

        # verify no more than one row was affected
        if rowcount > 1:
            # raise exception, bec/ we must not commit the transaction
            raise ValueError("More than one row was affected. Inserts: %s, Query: %s, Params: %s" % (rowcount, query, params))

        column_names = [desc[0] for desc in cursor.description]

        # Combine the column names with the row values into a dictionary
        row_dict = dict(zip(column_names, row))

        return row_dict

    @abstractmethod
    def _execute_orm_method(self, model_name: str, params: dict):
        raise NotImplementedError

    def fake_execute(self):
        pass

    def queries(self):
        pass


class CreateOrmExecutor(OrmExecutor):
    def _execute_orm_method(self, model_name: str, values: dict):
        new_record = self._orm.create(model_name, values)
        if not new_record:
            return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self._query, self._query_values, self.context)

        return ExecutionResult(self.line_number, self.operation_type, True, 1, self.table_name, self._query, self._query_values, self.context, new_record)


class UpdateOrmExecutor(OrmExecutor):
    def _execute_orm_method(self, model_name: str, record_id: int, values: dict):
        self._orm.update(model_name, record_id, values)


class DeleteOrmExecutor(OrmExecutor):
    def _execute_orm_method(self, model_name: str, record_id: int, values: dict):
        self._orm.delete(model_name, record_id)



"""
This class compiles a mapping into a list of parameters for each column.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from itertools import chain
from typing import List

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class ParametersRenderer:

    def render(self, mapping: Entity):
        # Return tuples because they are hashable, so we can use them as key in a dictionary
        return [tuple(self._attribute(a)) for a in mapping.attributes]

    def _attribute(self, attribute: AbstractAttribute) -> List[str]:
        if isinstance(attribute, Attribute):
            return [attribute.parameter]

        if isinstance(attribute, Reference):
            # recurse
            return self._attributes(attribute.attributes)

    def _attributes(self, attributes) -> List[str]:
        return list(chain(*[self._attribute(a) for a in attributes]))



"""
This abstract class declares methods to access the database model.

Author: Romke Jonker
Email: romke@stml.io
"""

from abc import ABC, abstractmethod


class ModelService(ABC):

    @abstractmethod
    def get_table(self, table_name):
        pass

    @abstractmethod
    def find_primary_keys(self, table):
        pass

    @abstractmethod
    def resolve_foreign_key_table(self, table, column_name):
        pass

    @abstractmethod
    def get_non_empty_columns(self, table):
        pass

    @abstractmethod
    def read_table(self, mapping: dict, where_clause=None):
        pass



"""
This class merges two mapping headers into a single mapping. The purpose is to let the user submit a mapping header and then
present him/her with a merged mapping header that contains the requested columns and the default columns in the correct order.

The requirements for the merged mapping are:
    - requested columns appear in order as requested
    - requested columns replace default columns, matched by attribute names
    - default columns appear in their own order, 'as much as possible'.
    - only requested columns are enabled


Author: Romke Jonker
Email: romke@rnadesign.net
"""
from stimula.stml.model import Entity, AbstractAttribute


class StmlMerger:
    def merge(self, default_mapping: Entity, requested_mapping: Entity) -> Entity:

        # verify both mappings have the same table
        assert default_mapping.name == requested_mapping.name

        # disable all default columns by removing 'in-use' attribute
        for a in default_mapping.attributes:
            a.enabled = False

        # map default columns by column key
        default_attributes_map = {a.name: a for a in default_mapping.attributes}

        # map requested columns by column key, but only for columns that have attributes
        requested_attributes_map = {a.name: a for a in requested_mapping.attributes if a}

        # merge requested columns into default columns
        result_attributes_map = self._merge_maps(default_attributes_map, requested_attributes_map, self._merge_attribute)

        # convert result column map to list
        result_attributes = list(result_attributes_map.values())

        return Entity(default_mapping.name, result_attributes)


    def _merge_attribute(self, default_attribute: AbstractAttribute, requested_attribute: AbstractAttribute) -> AbstractAttribute:
        # return requested column, but keep 'in-use' attribute from default column
        requested_attribute.in_use = default_attribute.in_use

        # also keep the 'primary-key' attribute from default column
        requested_attribute.primary_key = default_attribute.primary_key

        return requested_attribute

    def _merge_maps(self, a, b, f=lambda x, y: y):
        '''
        Sorts two dictionaries by their 'key' attribute.
        Resulting dictionary has:
        - all keys from a and b
        - values from b have precedence over values from a
        - order in b is preserved
        - order in a is preserved for the longest possible tail
        '''
        # get keys from b that do not exist in a

        # sort keys in a by their position in b
        b_keys = (list(b.keys()))
        i = len(b_keys) - 1
        c_keys = []
        for key in reversed(a.keys()):
            if key in c_keys:
                # already visited
                pass
            elif not key in b_keys:
                # not in b, simply add to c
                c_keys.insert(0, key)
            else:
                while i >= 0 and key != b_keys[i]:
                    # must still add, but keep items from b in order
                    c_keys.insert(0, b_keys[i])
                    i -= 1
                if i >= 0:
                    # found key in b, add to c
                    c_keys.insert(0, key)
                    i -= 1
        # get remaining keys from b
        remainders = b_keys[:i + 1]
        # insert remainders at the front of c_keys
        c_keys = remainders + c_keys

        # merge dictionaries, using lambda f to merge values that appear in both dictionaries
        merged = {**a, **b}
        for key in a.keys() & b.keys():
            merged[key] = f(a[key], b[key])

        # sort by c_keys
        result = {k: merged[k] for k in c_keys}

        return result



# sly/ast.py
import sys

class AST(object):
    
    @classmethod
    def __init_subclass__(cls, **kwargs):
        mod = sys.modules[cls.__module__]
        if not hasattr(cls, '__annotations__'):
            return

        hints = list(cls.__annotations__.items())

        def __init__(self, *args, **kwargs):
            if len(hints) != len(args):
                raise TypeError(f'Expected {len(hints)} arguments')
            for arg, (name, val) in zip(args, hints):
                if isinstance(val, str):
                    val = getattr(mod, val)
                if not isinstance(arg, val):
                    raise TypeError(f'{name} argument must be {val}')
                setattr(self, name, arg)

        cls.__init__ = __init__




"""
This class renders a mapping into a csv-style file header format

Author: Romke Jonker
Email: romke@stml.io
"""
from typing import List

from stimula.stml.model import Entity, AbstractAttribute, Reference, Attribute


class HeaderRenderer:
    # attributes to include as UI directives
    UI_ATTRIBUTES = ['enabled', 'primary_key', 'unique', 'in_use', 'default']
    # attributes to exclude from the modifiers list
    NON_MODIFIER_ATTRIBUTES = ['name', 'type', 'table', 'target_name', 'attributes', 'enabled', 'primary_key', 'foreign_key', 'in_use', 'default', 'parameter', 'alias', 'extension', 'deduplicate', 'orm_only', 'qualifier']

    def render_csv(self, mapping: Entity):
        # only return enabled columns
        return ', '.join([self._column(a) for a in mapping.attributes if a.enabled])

    def render_csv_unique(self, mapping: Entity):
        # only return unique columns
        return ', '.join([self._column(a) for a in mapping.attributes if a.unique])

    def render_list(self, mapping: Entity, enabled=False, include_skip=False, include_orm_only=False):
        # if enabled, only return enabled columns.
        # if include_skip, then include skip columns. This is useful when reading from CSV
        return [self._column(a) for a in mapping.attributes if (not a) or ((not enabled or a.enabled) and (include_skip or not a.skip) and (include_orm_only or not a.orm_only))]

    def render_list_unique(self, mapping: Entity):
        # return list of unique columns, skip empty columns
        return [self._column(a) for a in mapping.attributes if a and a.unique]

    def render_list_non_unique(self, mapping: Entity):
        # return list of non-unique columns
        # remove unique columns, also remove 'skip' columns because we don't want to write them to DB. Leave in 'orm-only' columns because we need them to write to ORM.
        # skip empty columns
        return [self._column(a) for a in mapping.attributes if a and (not a.unique) and (not a.skip)]

    def render_list_non_unique_root_extension(self, mapping: Entity):
        # return list of non-unique columns that are root extension columns
        # remove unique columns, also remove 'skip' columns because we don't want to write them to DB
        # skip empty columns
        columns = [self._column(a) for a in mapping.attributes if a and (not a.unique) and (not a.skip) and self._is_root_extension(a)]
        # remove empty columns
        return [c for c in columns if c]

    def _is_root_extension(self, attribute: AbstractAttribute):
        # return true if this is an extension column
        return isinstance(attribute, Reference) and attribute.extension

    def render_list_deduplicate(self, mapping: Entity):
        # return list of attributes to deduplicate. Skip empty columns.
        return [self._column(a) for a in mapping.attributes if a and a.deduplicate]

    def _column(self, attribute: AbstractAttribute):
        # empty column does not have attributes
        if not attribute:
            return ''
        attributes, _ = self._attributes([attribute])
        modifiers = self._modifiers(attribute)
        return attributes + modifiers

    def _attributes(self, attributes: List[AbstractAttribute]):
        keys, types = zip(*[self._attribute(a) for a in attributes])
        type = types[0] if len(set(types)) == 1 else None
        return ':'.join(keys), type

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            return attribute.name, attribute.type

        if isinstance(attribute, Reference):
            key, type = self._foreign_key(attribute)
            return attribute.name + key, type

    def _foreign_key(self, foreign_key: Reference):
        key, type = self._attributes(foreign_key.attributes)
        return f"({key})", type

    def _modifiers(self, attribute: AbstractAttribute):
        # convert to dict
        attribute_dict = attribute.to_dict()

        # use an exclusion list, so we can add more attributes later
        modifiers = [self._modifier(key, attribute_dict[key]) for key in sorted(attribute_dict.keys()) if key not in self.NON_MODIFIER_ATTRIBUTES]
        if not modifiers:
            return ''
        return '[' + ': '.join(modifiers) + ']'

    def _modifier(self, key, value):
        # replace '_' with '-' in key
        key = key.replace('_', '-')

        # if type is not yet string, then convert to string and lowercase, so we get 'true' and not 'True'
        if not isinstance(value, str):
            value = str(value).lower()

        # need to quote if value has any of the following: ' "\'='
        if any(c in value for c in [' ', '"', "'", '=']):
            return f'{key}="{value}"'
        return f'{key}={value}'



# docparse.py
#
# Support doc-string parsing classes

__all__ = [ 'DocParseMeta' ]

class DocParseMeta(type):
    '''
    Metaclass that processes the class docstring through a parser and
    incorporates the result into the resulting class definition. This
    allows Python classes to be defined with alternative syntax. 
    To use this class, you first need to define a lexer and parser:

        from sly import Lexer, Parser
        class MyLexer(Lexer):
           ...

        class MyParser(Parser):
           ...

    You then need to define a metaclass that inherits from DocParseMeta.
    This class must specify the associated lexer and parser classes.
    For example:

        class MyDocParseMeta(DocParseMeta):
            lexer = MyLexer
            parser = MyParser

    This metaclass is then used as a base for processing user-defined
    classes:

        class Base(metaclass=MyDocParseMeta):
            pass

        class Spam(Base):
            """
            doc string is parsed
            ...
            """

    It is expected that the MyParser() class would return a dictionary. 
    This dictionary is used to create the final class Spam in this example.
    '''

    @staticmethod
    def __new__(meta, clsname, bases, clsdict):
        if '__doc__' in clsdict:
            lexer = meta.lexer()
            parser = meta.parser()
            lexer.cls_name = parser.cls_name = clsname
            lexer.cls_qualname = parser.cls_qualname = clsdict['__qualname__']
            lexer.cls_module = parser.cls_module = clsdict['__module__']
            parsedict = parser.parse(lexer.tokenize(clsdict['__doc__']))
            assert isinstance(parsedict, dict), 'Parser must return a dictionary'
            clsdict.update(parsedict)
        return super().__new__(meta, clsname, bases, clsdict)

    @classmethod
    def __init_subclass__(cls):
        assert hasattr(cls, 'parser') and hasattr(cls, 'lexer')



from abc import ABC, abstractmethod

class AbstractORM(ABC):
    """Abstract class to define a generic ORM interface for CRUD operations."""

    @abstractmethod
    def create(self, model_name: str, values: dict):
        """Create a new record."""
        raise NotImplementedError

    @abstractmethod
    def read(self, model_name: str, record_id: int):
        """Read a record by its ID."""
        raise NotImplementedError

    @abstractmethod
    def update(self, model_name: str, record_id: int, values: dict):
        """Update a record with the given ID."""
        raise NotImplementedError

    @abstractmethod
    def delete(self, model_name: str, record_id: int):
        """Delete a record by its ID."""
        raise NotImplementedError



import requests

'''
This class reads documents from remote APIs.
'''


class ApiReader:
    def read_document(self, url_template, params):
        # get url, expand placeholders with values from row
        url = url_template.format(**params)

        # send http GET request to url
        response = requests.get(url)

        # assert that the response status code is 200
        assert response.status_code == 200, f"API call to {url} failed with status code {response.status_code}"

        return response.content
